#! /usr/bin/env ruby
# coding: utf-8

class Crysna::TransitionFinder::Cell < Crysna::Cell

  #@from_edges of a start cell must be set to be []
  attr_reader     :from_edges
  attr_reader     :reach_flag
  attr_reader     :atoms
  attr_accessor :name
  attr_accessor :energy

  class RangeError < Exception; end

  def initialize(sites, energy = nil, name = "")
    super(sites)
    @energy = energy
    @reach_flag = false
    @from_edges = [] # include Edge.
    @name = name
  end

  def ==(other)
    result = true
    result = false if @atoms            != other.atoms
    result = false if @energy           != other.energy
    result = false if @reach_flag != other.reach_flag
    result = false if @from_edges != other.from_edges
    result = false if @name             != other.name
    return result
  end

  def same_atoms?(other)
    return @atoms == other.atoms
  end

  def same_sites?(other)
    result = true
    @atoms.size.times do |index|
      if @atoms[index].site.name != other.atoms[index].site.name
        result = false
        break
      end
    end
    return result
  end

  #Migrate atom of index to site and in the cell
  #which is neighboring with the direction of global_vector.
  #Generating new cell which is not reached.
  #This method does not raise error even when destination site
  #is already occupied,
  #then, there are two atoms in the site.
  def migrate!(index, site, global_vector, io = File.open(File::NULL, "w"))
    io.puts "##{self.class}.migrate!(#{index}, #{site}, #{global_vector}"
    raise RangeError if index >= atoms.size

    atoms #sort
    @atoms[index].migrate!(site, global_vector)
    io.puts "Atoms after migration:"
    @atoms.each do |atom|
      io.puts "#{atom.element}, #{atom.site.name}, #{atom.site.global_vector}"
    end

    @atoms.sort!
    io.puts "Atoms after sort:"
    @atoms.each do |atom|
      io.puts "#{atom.element}, #{atom.site.name}, #{atom.site.global_vector}"
    end
    reset
  end

  #Not-destructive method of migrate!()
  def migrate(index, site, global_vector)
    result = Marshal.load(Marshal.dump(self))
    result.migrate!(index, site, global_vector)
    result
  end

  #Set reach flag to be true.
  #'edge' is the one between cells of already reached cell and self.
  def reach(edge)
    @reach_flag = true
    @from_edges << edge
  end

  #Return true if already reached.
  def reach?
    return @reach_flag
  end

  #Return true if all sites correspond to those of the other.
  #It is ok even if global vector is different.
  def periodically_equal?(other)
    self.atoms.sort.map{|atom| atom.site.name} == other.atoms.sort.map{|atom| atom.site.name}
  end

  def occupied_sitename
    result = ''
    result += atoms.map do |atom|
      periodic_num = ''
      atom.site.global_vector.each do |i|
        periodic_num += (i + 5).to_s
      end
      atom.site.name + periodic_num
    end . join "-"
    result
  end

  def inspect_short
    result = "#{self.class}: "
    result += occupied_sitename
    #result += '-'
    result += ", @energy = #{@energy}"
    result += ", @reach_flag = #{@reach_flag}"
    result += ", @from_edges = #{@from_edges}"
    result += ", @name = \"#{@name}\""
    return result
  end

  #def inspect
  #    result = 
  #    to_s
  #end
  #
  
  def operate(operation)
    result = super(operation)
    result.energy = @energy
    return result
  end

  private

  #Reset information of reach and edge.
  #(@reach_flag and @from_edges)
  def reset
    @reach_flag = false
    @from_edges = []
    @name = nil
  end

end

